==> bytedesk_kefu-0.1.5.tar.gz: lib/ui/chat/page/chat_im_page.dart <==
import 'dart:async';

import 'package:bytedesk_kefu/blocs/message_bloc/bloc.dart';
import 'package:bytedesk_kefu/blocs/thread_bloc/bloc.dart';
import 'package:bytedesk_kefu/model/messageProvider.dart';
import 'package:bytedesk_kefu/model/model.dart';
import 'package:bytedesk_kefu/mqtt/bytedesk_mqtt.dart';
import 'package:bytedesk_kefu/ui/chat/widget/message_widget.dart';
import 'package:bytedesk_kefu/ui/widget/image_choose_widget.dart';
import 'package:bytedesk_kefu/util/bytedesk_constants.dart';
import 'package:bytedesk_kefu/util/bytedesk_events.dart';
import 'package:flustars/flustars.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:image_picker/image_picker.dart';

// 此为：一对一、群聊、客服会话客服接待端 页面
class ChatIMPage extends StatefulWidget {
  //
  final String cid;
  final String title;
  final String custom;
  final String postscript;
  //
  ChatIMPage(
      {Key key,
      this.cid,
      this.title,
      this.custom,
      this.postscript})
      : super(key: key);
  //
  @override
  _ChatIMPageState createState() => _ChatIMPageState();
}

class _ChatIMPageState extends State<ChatIMPage>
    with
        AutomaticKeepAliveClientMixin<ChatIMPage>,
        TickerProviderStateMixin,
        WidgetsBindingObserver {
  //
  String _title;
  // 输入文字
  TextEditingController _textController = new TextEditingController();
  // 滚动监听
  ScrollController _scrollController = new ScrollController();
  // 聊天记录本地存储
  MessageProvider _messageProvider = new MessageProvider();
  // 聊天记录内存存储
  List<MessageWidget> _messages = <MessageWidget>[];
  // 图片
  ImagePicker _picker = ImagePicker();
  // 长连接
  BytedeskMqtt _bdMqtt = new BytedeskMqtt();
  // 当前用户uid
  String _currentUid = SpUtil.getString(BytedeskConstants.uid);
  // 当前会话
  Thread _currentThread;
  // 判断是否机器人对话状态
  // bool _isRobot;
  // 分页加载聊天记录
  int _page = 0;
  int _size = 20;
  //
  @override
  void initState() {
    super.initState();
    _title = widget.title;
    WidgetsBinding.instance.addObserver(this);
    _listener();
    setState(() {
      // _isRobot = false;
    });
  }

  //
  @override
  Widget build(BuildContext context) {
    super.build(context);
    //
    return Scaffold(
        appBar: AppBar(
          title: Text(_title),
          centerTitle: true,
          elevation: 0,
          actions: [
            // TODO: 客服会话：访客信息、转接会话
            // TODO: 一对一：个人信息

          ],
        ),
        body: MultiBlocListener(
            listeners: [
              BlocListener<ThreadBloc, ThreadState>(
                listener: (context, state) {
                  if (state is RequestThreadSuccess) {
                    setState(() {
                      // _isRobot = false;
                      _currentThread = state.threadResult.msg.thread;
                    });
                    // 插入本地
                    // _messageProvider.insert(state.threadResult.msg);
                    // 加载本地历史消息
                    _getMessages(_page, _size);
                    // _appendMessage(state.threadResult.msg);
                    //
                    if (state.threadResult.statusCode == 200 ||
                        state.threadResult.statusCode == 201) {
                      print('创建新会话');
                      // 发送商品信息
                      if (widget.custom != null &&
                          widget.custom.trim().length > 0) {
                        _bdMqtt.sendCommodityMessage(
                            widget.custom, _currentThread);
                      }
                      // 发送附言消息
                      if (widget.postscript != null &&
                          widget.postscript.trim().length > 0) {
                        _bdMqtt.sendTextMessage(
                            widget.postscript, _currentThread);
                      }
                    } else if (state.threadResult.statusCode == 202) {
                      print('提示排队中');
                      // 发送商品信息
                      if (widget.custom != null &&
                          widget.custom.trim().length > 0) {
                        _bdMqtt.sendCommodityMessage(
                            widget.custom, _currentThread);
                      }
                      // 发送附言消息
                      if (widget.postscript != null &&
                          widget.postscript.trim().length > 0) {
                        _bdMqtt.sendTextMessage(
                            widget.postscript, _currentThread);
                      }
                    } else if (state.threadResult.statusCode == 203) {
                      print('当前非工作时间，请自助查询或留言');
                      // TODO: 显示留言页面
                    } else if (state.threadResult.statusCode == 204) {
                      print('当前无客服在线，请自助查询或留言');
                      // TODO: 显示留言页面
                    } else if (state.threadResult.statusCode == 205) {
                      print('咨询前问卷');
                    } else if (state.threadResult.statusCode == 206) {
                      print('返回机器人初始欢迎语 + 欢迎问题列表');
                      setState(() {
                        // _isRobot = true;
                      });
                      // TODO: 显示问题列表
                    } else {
                      Fluttertoast.showToast(msg: "请求会话失败");
                    }
                  }
                },
              ),
              BlocListener<MessageBloc, MessageState>(
                listener: (context, state) {
                  print('message state change');
                  if (state is ReceiveMessageState) {
                    print('receive message:' + state.message.content);
                  } else if (state is UploadImageSuccess) {
                    _bdMqtt.sendImageMessage(state.url, _currentThread);
                  } else if (state is QueryAnswerSuccess) {
                  } else if (state is MessageAnswerSuccess) {
                    _appendMessage(state.query);
                    _appendMessage(state.answer);
                  } else if (state is RateAnswerSuccess) {}
                },
              ),
            ],
            child: Container(
              alignment: Alignment.bottomCenter,
              color: Color(0xFFDEEEEEE),
              child: Column(
                children: <Widget>[
                  // Expanded(
                  //   child: RefreshIndicator(
                  //     child: ListView.builder(
                  //       padding: EdgeInsets.all(8.0),
                  //       // reverse: true,
                  //       controller: _scrollController,
                  //       itemCount: _messages.length,
                  //       itemBuilder: (_, int index) => _messages[index],
                  //     ),
                  //     onRefresh: _loadMoreMessages,
                  //   ),
                  // ),
                  Expanded(
                    child: ListView.builder(
                      padding: new EdgeInsets.all(8.0),
                      reverse: true,
                      shrinkWrap: true,
                      controller: _scrollController,
                      itemBuilder: (_, int index) => _messages[index],
                      itemCount: _messages.length,
                    ),
                  ),
                  Divider(
                    height: 1.0,
                  ),
                  Container(
                    decoration: BoxDecoration(
                      color: Theme.of(context).cardColor,
                    ),
                    child: _textComposerWidget(),
                  ),
                ],
              ),
            )));
  }

  //
  Widget _textComposerWidget() {
    return IconTheme(
      data: IconThemeData(color: Colors.blue),
      child: Container(
        margin: const EdgeInsets.symmetric(horizontal: 8.0),
        child: Row(
          children: <Widget>[
            Container(
              // margin: const EdgeInsets.symmetric(horizontal: 4.0),
              child: new IconButton(
                icon: new Icon(Icons.add),
                onPressed: () {
                  // _getImage();
                  showModalBottomSheet(
                      context: context,
                      builder: (context) {
                        return ImageChooseWidget(
                          pickImageCallBack: () {
                            _pickImage();
                          },
                          takeImageCallBack: () {
                            _takeImage();
                          },
                        );
                      });
                },
              ),
            ),
            Flexible(
              child: TextField(
                onChanged: (value) {
                  if (value != null && value.trim().length > 0) {
                    _bdMqtt.sendPreviewMessage(value, _currentThread);
                  } else {
                    _bdMqtt.sendPreviewMessage('', _currentThread);
                  }
                },
                decoration: InputDecoration.collapsed(hintText: "输入内容..."),
                controller: _textController,
                onSubmitted: _handleSubmitted,
              ),
            ),
            Container(
              margin: const EdgeInsets.symmetric(horizontal: 4.0),
              child: IconButton(
                icon: Icon(Icons.send),
                onPressed: () => _handleSubmitted(_textController.text),
              ),
            )
          ],
        ),
      ),
    );
  }

  //
  @override
  bool get wantKeepAlive => true;

  //
  void _handleSubmitted(String text) {
    _textController.clear();
    //
    if (text.trim().length == 0) {
      return;
    }
    //
    _bdMqtt.sendTextMessage(text, _currentThread);
  }

  //
  _listener() {
    // 更新消息状态
    bytedeskEventBus.on<ReceiveMessageReceiptEventBus>().listen((event) {
      // print('更新状态:' + event.status);
      if (!this.mounted) {
        return;
      }
      // 更新界面, FIXME: 只有插入新消息，才会更新？
      for (var i = 0; i < _messages.length; i++) {
        MessageWidget messageWidget = _messages[i];
        if (messageWidget.message.mid == event.mid && 
          _messages[i].message.status != BytedeskConstants.MESSAGE_STATUS_READ) {
          setState(() {
            _messages[i].message.status = event.status;
          });
        }
      }
    });
    bytedeskEventBus.on<ReceiveMessagePreviewEventBus>().listen((event) {
      // print('消息预知');
      if (!this.mounted) {
        return;
      }
      setState(() {
        _title = '对方正在输入';
      });
      // 还原title
      Timer(
        Duration(seconds: 3),
        () {
          // print('timer');
          setState(() {
            _title = widget.title;
          });
        },
      );
    });
    // 同 DeleteMessageEventBus 事件
    // bytedeskEventBus.on<ReceiveMessageRecallEventBus>().listen((event) {
    //   print('消息撤回');
    // });
    // 接收到新消息
    bytedeskEventBus.on<ReceiveMessageEventBus>().listen((event) {
      // print('receive message:' + event.message.content);
      if (_currentThread != null &&
          (event.message.thread.topic != _currentThread.topic)) {
        return;
      }
      // 非自己发送的，发送已读回执
      if (event.message.isSend == 0) {
        _bdMqtt.sendReceiptReadMessage(event.message.mid, event.message.thread);
      }
      // 界面显示
      MessageWidget messageWidget = new MessageWidget(
          message: event.message,
          animationController: new AnimationController(
              vsync: this, duration: Duration(milliseconds: 500)));
      if (this.mounted) {
        setState(() {
          _messages.insert(0, messageWidget);
          // _messages.add(messageWidget);
        });
      }
    });
    // 删除消息
    bytedeskEventBus.on<DeleteMessageEventBus>().listen((event) {
      //
      if (this.mounted) {
        // 从sqlite中删除
        _messageProvider.delete(event.mid);
        // 更新界面
        setState(() {
          _messages.removeWhere((element) => element.message.mid == event.mid);
        });
      }
    });
    // 滚动监听, https://learnku.com/articles/30338
    _scrollController.addListener(() {
      // 隐藏软键盘
      FocusScope.of(context).requestFocus(FocusNode());
      // 如果滑动到底部
      if (_scrollController.position.pixels ==
          _scrollController.position.maxScrollExtent) {}
    });
  }

  // 选择图片
  Future<void> _pickImage() async {
    try {
      PickedFile pickedFile = await _picker.getImage(
          source: ImageSource.gallery, maxWidth: 800, imageQuality: 95);
      print('pick image path: ${pickedFile.path}');
      //
      BlocProvider.of<MessageBloc>(context)
        ..add(UploadImageEvent(filePath: pickedFile.path));
    } catch (e) {
      print('pick image error ${e.toString()}');
      Fluttertoast.showToast(msg: "未选取图片");
    }
  }

  // 拍照
  Future<void> _takeImage() async {
    try {
      PickedFile pickedFile = await _picker.getImage(
          source: ImageSource.camera, maxWidth: 800, imageQuality: 95);
      print('take image path: ${pickedFile.path}');
      //
      BlocProvider.of<MessageBloc>(context)
        ..add(UploadImageEvent(filePath: pickedFile.path));
    } catch (e) {
      print('take image error ${e.toString()}');
      Fluttertoast.showToast(msg: "未选取图片");
    }
  }

  // 加载更多聊天记录
  Future<void> _loadMoreMessages() async {
    print('load more');
    // TODO: 从服务器加载
    _getMessages(_page, _size);
  }

  // 分页加载历史聊天记录
  Future<Null> _getMessages(int page, int size) async {
    //
    List<Message> messageList = await _messageProvider.getTopicMessages(
        _currentThread.topic, _currentUid, page, size);
    messageList.forEach((message) {
      MessageWidget messageWidget = new MessageWidget(
          message: message,
          animationController: new AnimationController(
              vsync: this, duration: Duration(milliseconds: 500)));
      if (this.mounted) {
        setState(() {
          // _messages.insert(0, messageWidget);
          _messages.add(messageWidget);
        });
      }
    });
    //
    _page += 1;
  }

  Future<Null> _appendMessage(Message message) async {
    print('append:' + message.mid);
    //
    _messageProvider.insert(message);
    MessageWidget messageWidget = new MessageWidget(
        message: message,
        animationController: new AnimationController(
            vsync: this, duration: Duration(milliseconds: 500)));
    // 防止：Unhandled Exception: setState() called after dispose()，添加判断
    if (this.mounted) {
      setState(() {
        _messages.insert(0, messageWidget);
      });
    }
  }

  void scrollToBottom() {
    // After 1 second, it takes you to the bottom of the ListView
    // Timer(
    //   Duration(seconds: 1),
    //   () => _scrollController.jumpTo(_scrollController.position.maxScrollExtent),
    // );
    _scrollController.animateTo(
      _scrollController.position.maxScrollExtent,
      duration: Duration(seconds: 1),
      curve: Curves.fastOutSlowIn,
    );
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    print("didChangeAppLifecycleState:" + state.toString());
    switch (state) {
      case AppLifecycleState.inactive: // 处于这种状态的应用程序应该假设它们可能在任何时候暂停。
        break;
      case AppLifecycleState.paused: // 应用程序不可见，后台
        break;
      case AppLifecycleState.resumed: // 应用程序可见，前台
        // APP切换到前台之后，重连
        // BytedeskUtils.mqttReConnect();
        // TODO: 拉取离线消息
        break;
      case AppLifecycleState.detached: // 申请将暂时暂停
        break;
    }
  }

  @override
  void dispose() {
    WidgetsBinding.instance.removeObserver(this);
    super.dispose();
  }
}

==> bytedesk_kefu-0.2.6.tar.gz: lib/ui/chat/page/chat_im_page.dart <==
import 'dart:async';

import 'package:bytedesk_kefu/blocs/message_bloc/bloc.dart';
import 'package:bytedesk_kefu/blocs/thread_bloc/bloc.dart';
import 'package:bytedesk_kefu/model/messageProvider.dart';
import 'package:bytedesk_kefu/model/model.dart';
import 'package:bytedesk_kefu/mqtt/bytedesk_mqtt.dart';
import 'package:bytedesk_kefu/ui/chat/widget/message_widget.dart';
import 'package:bytedesk_kefu/ui/widget/image_choose_widget.dart';
import 'package:bytedesk_kefu/util/bytedesk_constants.dart';
import 'package:bytedesk_kefu/util/bytedesk_events.dart';
import 'package:flustars/flustars.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:image_picker/image_picker.dart';

// 此为：一对一、群聊、客服会话客服接待端 页面
class ChatIMPage extends StatefulWidget {
  //
  final String cid;
  final String title;
  final String custom;
  final String postscript;
  //
  ChatIMPage(
      {Key key,
      this.cid,
      this.title,
      this.custom,
      this.postscript})
      : super(key: key);
  //
  @override
  _ChatIMPageState createState() => _ChatIMPageState();
}

class _ChatIMPageState extends State<ChatIMPage>
    with
        AutomaticKeepAliveClientMixin<ChatIMPage>,
        TickerProviderStateMixin,
        WidgetsBindingObserver {
  //
  String _title;
  // 输入文字
  TextEditingController _textController = new TextEditingController();
  // 滚动监听
  ScrollController _scrollController = new ScrollController();
  // 聊天记录本地存储
  MessageProvider _messageProvider = new MessageProvider();
  // 聊天记录内存存储
  List<MessageWidget> _messages = <MessageWidget>[];
  // 图片
  ImagePicker _picker = ImagePicker();
  // 长连接
  BytedeskMqtt _bdMqtt = new BytedeskMqtt();
  // 当前用户uid
  String _currentUid = SpUtil.getString(BytedeskConstants.uid);
  // 当前会话
  Thread _currentThread;
  // 判断是否机器人对话状态
  // bool _isRobot;
  // 分页加载聊天记录
  int _page = 0;
  int _size = 20;
  //
  @override
  void initState() {
    super.initState();
    _title = widget.title;
    WidgetsBinding.instance.addObserver(this);
    _listener();
    setState(() {
      // _isRobot = false;
    });
  }

  //
  @override
  Widget build(BuildContext context) {
    super.build(context);
    //
    return Scaffold(
        appBar: AppBar(
          title: Text(_title),
          centerTitle: true,
          elevation: 0,
          actions: [
            // TODO: 客服会话：访客信息、转接会话
            // TODO: 一对一：个人信息

          ],
        ),
        body: MultiBlocListener(
            listeners: [
              BlocListener<ThreadBloc, ThreadState>(
                listener: (context, state) {
                  if (state is RequestThreadSuccess) {
                    setState(() {
                      // _isRobot = false;
                      _currentThread = state.threadResult.msg.thread;
                    });
                    // 插入本地
                    // _messageProvider.insert(state.threadResult.msg);
                    // 加载本地历史消息
                    _getMessages(_page, _size);
                    // _appendMessage(state.threadResult.msg);
                    //
                    if (state.threadResult.statusCode == 200 ||
                        state.threadResult.statusCode == 201) {
                      print('创建新会话');
                      // 发送商品信息
                      if (widget.custom != null &&
                          widget.custom.trim().length > 0) {
                        _bdMqtt.sendCommodityMessage(
                            widget.custom, _currentThread);
                      }
                      // 发送附言消息
                      if (widget.postscript != null &&
                          widget.postscript.trim().length > 0) {
                        _bdMqtt.sendTextMessage(
                            widget.postscript, _currentThread);
                      }
                    } else if (state.threadResult.statusCode == 202) {
                      print('提示排队中');
                      // 发送商品信息
                      if (widget.custom != null &&
                          widget.custom.trim().length > 0) {
                        _bdMqtt.sendCommodityMessage(
                            widget.custom, _currentThread);
                      }
                      // 发送附言消息
                      if (widget.postscript != null &&
                          widget.postscript.trim().length > 0) {
                        _bdMqtt.sendTextMessage(
                            widget.postscript, _currentThread);
                      }
                    } else if (state.threadResult.statusCode == 203) {
                      print('当前非工作时间，请自助查询或留言');
                      // TODO: 显示留言页面
                    } else if (state.threadResult.statusCode == 204) {
                      print('当前无客服在线，请自助查询或留言');
                      // TODO: 显示留言页面
                    } else if (state.threadResult.statusCode == 205) {
                      print('咨询前问卷');
                    } else if (state.threadResult.statusCode == 206) {
                      print('返回机器人初始欢迎语 + 欢迎问题列表');
                      setState(() {
                        // _isRobot = true;
                      });
                      // TODO: 显示问题列表
                    } else {
                      Fluttertoast.showToast(msg: "请求会话失败");
                    }
                  }
                },
              ),
              BlocListener<MessageBloc, MessageState>(
                listener: (context, state) {
                  print('message state change');
                  if (state is ReceiveMessageState) {
                    print('receive message:' + state.message.content);
                  } else if (state is UploadImageSuccess) {
                    _bdMqtt.sendImageMessage(state.url, _currentThread);
                  } else if (state is QueryAnswerSuccess) {
                  } else if (state is MessageAnswerSuccess) {
                    _appendMessage(state.query);
                    _appendMessage(state.answer);
                  } else if (state is RateAnswerSuccess) {}
                },
              ),
            ],
            child: Container(
              alignment: Alignment.bottomCenter,
              color: Color(0xFFDEEEEEE),
              child: Column(
                children: <Widget>[
                  // Expanded(
                  //   child: RefreshIndicator(
                  //     child: ListView.builder(
                  //       padding: EdgeInsets.all(8.0),
                  //       // reverse: true,
                  //       controller: _scrollController,
                  //       itemCount: _messages.length,
                  //       itemBuilder: (_, int index) => _messages[index],
                  //     ),
                  //     onRefresh: _loadMoreMessages,
                  //   ),
                  // ),
                  Expanded(
                    child: ListView.builder(
                      padding: new EdgeInsets.all(8.0),
                      reverse: true,
                      shrinkWrap: true,
                      controller: _scrollController,
                      itemBuilder: (_, int index) => _messages[index],
                      itemCount: _messages.length,
                    ),
                  ),
                  Divider(
                    height: 1.0,
                  ),
                  Container(
                    decoration: BoxDecoration(
                      color: Theme.of(context).cardColor,
                    ),
                    child: _textComposerWidget(),
                  ),
                ],
              ),
            )));
  }

  //
  Widget _textComposerWidget() {
    return IconTheme(
      data: IconThemeData(color: Colors.blue),
      child: Container(
        margin: const EdgeInsets.symmetric(horizontal: 8.0),
        child: Row(
          children: <Widget>[
            Container(
              // margin: const EdgeInsets.symmetric(horizontal: 4.0),
              child: new IconButton(
                icon: new Icon(Icons.add),
                onPressed: () {
                  // _getImage();
                  showModalBottomSheet(
                      context: context,
                      builder: (context) {
                        return ImageChooseWidget(
                          pickImageCallBack: () {
                            _pickImage();
                          },
                          takeImageCallBack: () {
                            _takeImage();
                          },
                        );
                      });
                },
              ),
            ),
            Flexible(
              child: TextField(
                onChanged: (value) {
                  if (value != null && value.trim().length > 0) {
                    _bdMqtt.sendPreviewMessage(value, _currentThread);
                  } else {
                    _bdMqtt.sendPreviewMessage('', _currentThread);
                  }
                },
                decoration: InputDecoration.collapsed(hintText: "输入内容..."),
                controller: _textController,
                onSubmitted: _handleSubmitted,
              ),
            ),
            Container(
              margin: const EdgeInsets.symmetric(horizontal: 4.0),
              child: IconButton(
                icon: Icon(Icons.send),
                onPressed: () => _handleSubmitted(_textController.text),
              ),
            )
          ],
        ),
      ),
    );
  }

  //
  @override
  bool get wantKeepAlive => true;

  //
  void _handleSubmitted(String text) {
    _textController.clear();
    //
    if (text.trim().length == 0) {
      return;
    }
    //
    _bdMqtt.sendTextMessage(text, _currentThread);
  }

  //
  _listener() {
    // 更新消息状态
    bytedeskEventBus.on<ReceiveMessageReceiptEventBus>().listen((event) {
      // print('更新状态:' + event.status);
      if (!this.mounted) {
        return;
      }
      // 更新界面, FIXME: 只有插入新消息，才会更新？
      for (var i = 0; i < _messages.length; i++) {
        MessageWidget messageWidget = _messages[i];
        if (messageWidget.message.mid == event.mid && 
          _messages[i].message.status != BytedeskConstants.MESSAGE_STATUS_READ) {
          setState(() {
            _messages[i].message.status = event.status;
          });
        }
      }
    });
    bytedeskEventBus.on<ReceiveMessagePreviewEventBus>().listen((event) {
      // print('消息预知');
      if (!this.mounted) {
        return;
      }
      setState(() {
        _title = '对方正在输入';
      });
      // 还原title
      Timer(
        Duration(seconds: 3),
        () {
          // print('timer');
          setState(() {
            _title = widget.title;
          });
        },
      );
    });
    // 同 DeleteMessageEventBus 事件
    // bytedeskEventBus.on<ReceiveMessageRecallEventBus>().listen((event) {
    //   print('消息撤回');
    // });
    // 接收到新消息
    bytedeskEventBus.on<ReceiveMessageEventBus>().listen((event) {
      // print('receive message:' + event.message.content);
      if (_currentThread != null &&
          (event.message.thread.topic != _currentThread.topic)) {
        return;
      }
      // 非自己发送的，发送已读回执
      if (event.message.isSend == 0) {
        _bdMqtt.sendReceiptReadMessage(event.message.mid, event.message.thread);
      }
      // 界面显示
      MessageWidget messageWidget = new MessageWidget(
          message: event.message,
          animationController: new AnimationController(
              vsync: this, duration: Duration(milliseconds: 500)));
      if (this.mounted) {
        setState(() {
          _messages.insert(0, messageWidget);
          // _messages.add(messageWidget);
        });
      }
    });
    // 删除消息
    bytedeskEventBus.on<DeleteMessageEventBus>().listen((event) {
      //
      if (this.mounted) {
        // 从sqlite中删除
        _messageProvider.delete(event.mid);
        // 更新界面
        setState(() {
          _messages.removeWhere((element) => element.message.mid == event.mid);
        });
      }
    });
    // 滚动监听, https://learnku.com/articles/30338
    _scrollController.addListener(() {
      // 隐藏软键盘
      FocusScope.of(context).requestFocus(FocusNode());
      // 如果滑动到底部
      if (_scrollController.position.pixels ==
          _scrollController.position.maxScrollExtent) {}
    });
  }

  // 选择图片
  Future<void> _pickImage() async {
    try {
      PickedFile pickedFile = await _picker.getImage(
          source: ImageSource.gallery, maxWidth: 800, imageQuality: 95);
      print('pick image path: ${pickedFile.path}');
      //
      BlocProvider.of<MessageBloc>(context)
        ..add(UploadImageEvent(filePath: pickedFile.path));
    } catch (e) {
      print('pick image error ${e.toString()}');
      Fluttertoast.showToast(msg: "未选取图片");
    }
  }

  // 拍照
  Future<void> _takeImage() async {
    try {
      PickedFile pickedFile = await _picker.getImage(
          source: ImageSource.camera, maxWidth: 800, imageQuality: 95);
      print('take image path: ${pickedFile.path}');
      //
      BlocProvider.of<MessageBloc>(context)
        ..add(UploadImageEvent(filePath: pickedFile.path));
    } catch (e) {
      print('take image error ${e.toString()}');
      Fluttertoast.showToast(msg: "未选取图片");
    }
  }

  // 加载更多聊天记录
  // Future<void> _loadMoreMessages() async {
  //   print('load more');
  //   // TODO: 从服务器加载
  //   _getMessages(_page, _size);
  // }

  // 分页加载历史聊天记录
  Future<Null> _getMessages(int page, int size) async {
    //
    List<Message> messageList = await _messageProvider.getTopicMessages(
        _currentThread.topic, _currentUid, page, size);
    messageList.forEach((message) {
      MessageWidget messageWidget = new MessageWidget(
          message: message,
          animationController: new AnimationController(
              vsync: this, duration: Duration(milliseconds: 500)));
      if (this.mounted) {
        setState(() {
          // _messages.insert(0, messageWidget);
          _messages.add(messageWidget);
        });
      }
    });
    //
    _page += 1;
  }

  Future<Null> _appendMessage(Message message) async {
    print('append:' + message.mid);
    //
    _messageProvider.insert(message);
    MessageWidget messageWidget = new MessageWidget(
        message: message,
        animationController: new AnimationController(
            vsync: this, duration: Duration(milliseconds: 500)));
    // 防止：Unhandled Exception: setState() called after dispose()，添加判断
    if (this.mounted) {
      setState(() {
        _messages.insert(0, messageWidget);
      });
    }
  }

  void scrollToBottom() {
    // After 1 second, it takes you to the bottom of the ListView
    // Timer(
    //   Duration(seconds: 1),
    //   () => _scrollController.jumpTo(_scrollController.position.maxScrollExtent),
    // );
    _scrollController.animateTo(
      _scrollController.position.maxScrollExtent,
      duration: Duration(seconds: 1),
      curve: Curves.fastOutSlowIn,
    );
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    print("didChangeAppLifecycleState:" + state.toString());
    switch (state) {
      case AppLifecycleState.inactive: // 处于这种状态的应用程序应该假设它们可能在任何时候暂停。
        break;
      case AppLifecycleState.paused: // 应用程序不可见，后台
        break;
      case AppLifecycleState.resumed: // 应用程序可见，前台
        // APP切换到前台之后，重连
        // BytedeskUtils.mqttReConnect();
        // TODO: 拉取离线消息
        break;
      case AppLifecycleState.detached: // 申请将暂时暂停
        break;
    }
  }

  @override
  void dispose() {
    WidgetsBinding.instance.removeObserver(this);
    super.dispose();
  }
}
